/***************************************************************************
Copyright (c) 2006 Peter Korzuszek, Damian Rakowski, KateOS Project
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 
*Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
*Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE. 
****************************************************************************/

#ifndef ST_SHARED_MEMORY_H
#define ST_SHARED_MEMORY_H

#include <stdlib.h>
#include <string>
#include <stdio.h>

#include <assert.h>

//static void _null( const char*, ... ){}

//#define _printf _null

class StSharedMemory
{
	public:
		StSharedMemory() { init(); lock(); }
		StSharedMemory( const StSharedMemory& copy );
		~StSharedMemory() { unlock(); clean(); }
		
		/** Malloc new memory */
		void malloc( unsigned size );
		
		/** Resize the memory */
		void realloc( unsigned size );
		
		/** Free memory */
		void free();
		
		/*! Detach from data and get ready to renew the instance */
		void renew();
			
		void* getPointer() const { return pointer; }
		unsigned getMemorySize() const { return memorySize; }
		unsigned getInstances() const { return *instances; }
		
		/*!
			This function is DANGEROUS if you don't know what you're doing.
			Use it only when memory was reallocated outside this class.
		*/
		void setMemorySize( unsigned size ){ memorySize = size; }
		
		/*!
			This function is DANGEROUS of you don't know what you're doing.
			Use only when memory is not allocated.
		*/
		void setMemoryPointer( void* pointer ){ this->pointer = pointer; }
		
		/*! If data should no longer be shared with other instances, use this function*/
		void makeCopy();
		
		char& operator[]( unsigned index ){ return ((char*)pointer)[index]; }
		const char& operator[]( unsigned index ) const { return ((char*)pointer)[index]; }
		StSharedMemory& operator=( const StSharedMemory& copy );
	
	private:
		unsigned* instances;
		unsigned memorySize;
		void* pointer;
		
		/*! If no instances are left memory will be freed. */
		void clean();
		
		void init();
		
		// carefull with these!
		void lock() {  /*_printf("shared (%p): instance++: ", this);*/ (*instances)++; /*_printf("%d\n", *instances);*/ }
		void unlock() { assert( *instances > 0 ); /* _printf("shared (%p): instance--: ", this); */ (*instances)--; /*_printf("%d\n", *instances);*/ }
};

// --------------------------------------------------------

inline StSharedMemory::StSharedMemory( const StSharedMemory& copy )
{
	//_printf("shared (%p): copy constructor from %p\n", this, &copy);
	
	instances = copy.instances;
	memorySize = copy.memorySize;
	pointer = copy.pointer;
	
	lock();
}

// --------------------------------------------------------

inline StSharedMemory& StSharedMemory::operator=( const StSharedMemory& copy )
{
	unlock();
	clean();
	
	//_printf("shared (%p): copy from %p\n", this, &copy);
	
	instances = copy.instances;
	memorySize = copy.memorySize;
	pointer = copy.pointer;
	
	lock();
	
	return *this;
}

// --------------------------------------------------------

inline void StSharedMemory::clean()
{
	if( *instances == 0 ){
		//_printf("shared (%p): cleaning\n", this);
		if( instances != NULL ){
			::free( (void*)instances );
			instances = NULL;
		}
		
		if( pointer != NULL ){
			::free( pointer );
			pointer = NULL;
		}
		
		memorySize = 0;
	}
}

// --------------------------------------------------------

inline void StSharedMemory::init()
{
	//_printf("shared (%p): initialising\n", this);
	
	instances = (unsigned*)::malloc(sizeof(unsigned));
	
	memorySize = 0;
	*instances = 0;
	pointer = NULL;
}

// --------------------------------------------------------

inline void StSharedMemory::makeCopy()
{
	if( *instances <= 1 || pointer == NULL )
		return;
	
	//_printf("shared (%p): makeCopy()\n", this);
	
	unlock();
	
	void* npointer = ::malloc( memorySize );
	memcpy( npointer, pointer, memorySize );
	pointer = npointer;
	
	instances = (unsigned*)::malloc( sizeof(unsigned) );
	*instances=0;
	
	lock();
}

// --------------------------------------------------------

inline void StSharedMemory::malloc( unsigned size )
{	
	free();
	
	//_printf("shared (%p): malloc(%u)\n", this, size);
	
	pointer = ::malloc(size);
	memorySize=size;
}

// --------------------------------------------------------

inline void StSharedMemory::realloc( unsigned size )
{
	makeCopy();
	
	//_printf("shared (%p): realloc(%u)\n", this, size);
	
	pointer = ::realloc( pointer, size );
	memorySize=size;
}

// --------------------------------------------------------

inline void StSharedMemory::free()
{
	makeCopy();
	
	//_printf("shared (%p): free()\n", this);
	
	memorySize = 0;
	::free( pointer );
	pointer = NULL;
}

// --------------------------------------------------------

inline void StSharedMemory::renew()
{
	unlock();
	clean();
	
	init();
	lock();
}

#endif //ST_SHARED_MEMORY_H
